/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2019, ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _TMIV_VPCCBITSTREAM_ATLASFRAMEPARAMETERSETRBSP_H_
#define _TMIV_VPCCBITSTREAM_ATLASFRAMEPARAMETERSETRBSP_H_

#include <TMIV/Common/Bitstream.h>
#include <TMIV/VpccBitstream/AtlasSequenceParameterSetRBSP.h>

#include <cstdint>
#include <cstdlib>
#include <iosfwd>

namespace TMIV::VpccBitstream {
// 23090-5: atlas_frame_tile_information()
class AtlasFrameTileInformation {
public:
  friend auto operator<<(std::ostream &stream, const AtlasFrameTileInformation &x)
      -> std::ostream &;

  constexpr auto afti_single_tile_in_atlas_frame_flag() const noexcept { return true; }

  auto operator==(const AtlasFrameTileInformation &) const noexcept -> bool { return true; }
  auto operator!=(const AtlasFrameTileInformation &) const noexcept -> bool { return false; }

  static auto decodeFrom(Common::InputBitstream &bitstream) -> AtlasFrameTileInformation;

  static void encodeTo(Common::OutputBitstream &stream);
};

// 23090-5: atlas_frame_parameter_set_rbsp()
class AtlasFrameParameterSetRBSP {
public:
  friend auto operator<<(std::ostream &stream, const AtlasFrameParameterSetRBSP &x)
      -> std::ostream &;

  constexpr auto afps_atlas_frame_parameter_set_id() const noexcept;
  constexpr auto afps_atlas_sequence_parameter_set_id() const noexcept;
  constexpr auto atlas_frame_tile_information() const noexcept;
  constexpr auto afps_num_ref_idx_default_active() const noexcept;
  constexpr auto afps_additional_lt_afoc_lsb_len() const noexcept;
  constexpr auto afps_2d_pos_x_bit_count() const noexcept;
  constexpr auto afps_2d_pos_y_bit_count() const noexcept;
  constexpr auto afps_3d_pos_x_bit_count() const noexcept;
  constexpr auto afps_3d_pos_y_bit_count() const noexcept;
  constexpr auto afps_lod_bit_count() const noexcept;
  constexpr auto afps_override_eom_for_depth_flag() const noexcept;
  constexpr auto afps_raw_3d_pos_bit_count_explicit_mode_flag() const noexcept;
  constexpr auto afps_extension_present_flag() const noexcept;

  constexpr auto &afps_atlas_frame_parameter_set_id(const std::uint8_t value) noexcept;
  constexpr auto &afps_atlas_sequence_parameter_set_id(const std::uint8_t value) noexcept;
  constexpr auto &atlas_frame_tile_information(const AtlasFrameTileInformation &value) noexcept;
  constexpr auto &afps_num_ref_idx_default_active(const std::uint8_t value) noexcept;
  constexpr auto &afps_additional_lt_afoc_lsb_len(const std::uint8_t value) noexcept;
  constexpr auto &afps_2d_pos_x_bit_count(const std::uint8_t value) noexcept;
  constexpr auto &afps_2d_pos_y_bit_count(const std::uint8_t value) noexcept;
  constexpr auto &afps_3d_pos_x_bit_count(const std::uint8_t value) noexcept;
  constexpr auto &afps_3d_pos_y_bit_count(const std::uint8_t value) noexcept;
  constexpr auto &afps_lod_bit_count(const std::uint8_t value) noexcept;
  constexpr auto &afps_override_eom_for_depth_flag(const bool value) noexcept;
  constexpr auto &afps_raw_3d_pos_bit_count_explicit_mode_flag(const bool value) noexcept;
  constexpr auto &afps_extension_present_flag(const bool value) noexcept;

  auto operator==(const AtlasFrameParameterSetRBSP &other) const noexcept -> bool;
  auto operator!=(const AtlasFrameParameterSetRBSP &other) const noexcept -> bool;

  static auto decodeFrom(std::istream &stream, const AtlasSequenceParameterSetRBSP &asps)
      -> AtlasFrameParameterSetRBSP;

  void encodeTo(std::ostream &stream, const AtlasSequenceParameterSetRBSP &asps) const;

private:
  std::uint8_t m_afps_atlas_frame_parameter_set_id;
  std::uint8_t m_afps_atlas_sequence_parameter_set_id;
  AtlasFrameTileInformation m_atlas_frame_tile_information;
  std::uint8_t m_afps_num_ref_idx_default_active;
  std::uint8_t m_afps_additional_lt_afoc_lsb_len;
  std::uint8_t m_afps_2d_pos_x_bit_count;
  std::uint8_t m_afps_2d_pos_y_bit_count;
  std::uint8_t m_afps_3d_pos_x_bit_count;
  std::uint8_t m_afps_3d_pos_y_bit_count;
  std::uint8_t m_afps_lod_bit_count;
  bool m_afps_override_eom_for_depth_flag;
  bool m_afps_raw_3d_pos_bit_count_explicit_mode_flag;
  bool m_afps_extension_present_flag;
};
} // namespace TMIV::VpccBitstream

#include "AtlasFrameParameterSetRBSP.hpp"

#endif
